#include "DerivedFunctions.h"
#include "CompositeGridOperators.h"
#include "ParallelUtility.h"

//\begin{>>DerivedFunctionsInclude.tex}{\subsection{setupUserDefinedDerivedFunction}} 
int DerivedFunctions::
setupUserDefinedDerivedFunction(GenericGraphicsInterface & gi, 
                                int numberOfComponents, 
				aString *componentNames )
//==============================================================================================
// /Description:
//    User defined derived functions. This function is used to setup and define a derived function.
// The function getUserDefinedDerivedFunction is called to actually evaluate the derived function.
// Choose the "user defined" option from the DerivedFunctions::update options to have this routine
// called.
// /Notes:
//  \begin{itemize}
//    \item To to add a new user defined derived-function edit this file (Ogshow/userDefinedDerivedFunction.C)
//          and rebuild the plotStuff executable using the new version. This can be done by rebuilding the
//           Overture library in place, or by copying the files plotStuffDriver.C, plotStuff.C and 
//           userDefinedDerivedFunction.C to a new location and building a separate version of plotStuff.
//  \end{itemize}
//
// /Return values: 0=success, non-zero=failure.
//\end{DerivedFunctionsInclude.tex}
//==============================================================================================

{
  assert( showFileReader!=NULL );
  ShowFileReader & reader = *showFileReader;

  aString menu[]=
  {
    "!User Defined Derived Functions",
    "gas mass fraction",
    "mixture density",
    "mixture schlieren",
    "scaled schlieren",
    "mixture temperature",
    "reaction rate",
    "masked progress",
    "material interface",
    "stressNorm",
    "shearStress",
    "exit",
    ""
  };
  aString answer;
  
  // *** make a list of component names for later use ***
  aString *cNames = new aString [numberOfComponents+1];
  int n;
  for( n=0; n<numberOfComponents; n++ )
  { 
    cNames[n]=componentNames[n]; 
  }
  n=numberOfComponents;
  cNames[n]="";            // null terminated list

  // list of species 
  int numberOfSpecies=0;
  reader.getGeneralParameter("numberOfSpecies",numberOfSpecies );
  assert( numberOfSpecies>=0 );
  
  int sc=-1;
  if( numberOfSpecies>=0 )
    reader.getGeneralParameter("speciesComponent",sc);
  else
    sc=0;

  // *** make a list of species for later use ***
  aString *speciesNames = new aString [numberOfSpecies+1];
  for( n=0; n<numberOfSpecies; n++ )
  { 
    speciesNames[n]=componentNames[sc+n]; 
  }
  n=numberOfSpecies;
  speciesNames[n]=""; // null terminated list

  for( ;; )
  {
    int choice = gi.getMenuItem(menu,answer,"choose an option");
    
    if( answer=="exit" || answer=="done" )
    {
      break;
    }
    else if( answer=="gas mass fraction" )
    {
      add( userDefined,"gas mass fraction" );
      printF("Gas mass fraction was chosen.\n");
    }
    else if( answer=="mixture density" )
    {
      add( userDefined,"mixture density" );
      printF("Mixture density was chosen.\n");
    }
    else if( answer=="mixture schlieren" )
    {
      add( userDefined,"mixture schlieren" );
      printF("Mixture schlieren was chosen.\n");
    }
    else if( answer=="scaled schlieren" )
    {
      add( userDefined,"scaled schlieren" );
      printF("Scaled schlieren was chosen.\n");
    }
    else if( answer=="mixture temperature" )
    {
      add( userDefined,"mixture temperature" );
      printF("Mixture temperature was chosen.\n");
    }
    else if( answer=="reaction rate" )
    {
      int species=0;

      // we query to determine the species for which the reaction rate is required
      aString answer2;
      int s = gi.getMenuItem(speciesNames,answer2,"reaction rate for which species?");
      if( s>=0 && s<numberOfSpecies )
      {
	species=s+sc;
      }
      else
      {
	cout << "Unknown response: [" << answer << "]\n";
	gi.stopReadingCommandFile();
      }

      // save a user defined function with a given name ("reaction rate") 
      //    and an optional integer identifier (species)
      add( userDefined,"reaction rate",species );

      printF("Species %s was chosen (component number %i).\n",(const char*)componentNames[species],species);

    }
    else if( answer=="masked progress" )
    {
      if( numberOfSpecies<2 )
      {
	cout << "Need two species variables [lambda,mu]\n";
	gi.stopReadingCommandFile();
      }
      else
      {
	add( userDefined,"masked progress" );
	printF("Masked progress was chosen.\n");
      }
    }
    else if( answer=="material interface" )
    {
      if( numberOfSpecies<2 )
      {
	cout << "Need two species variables [lambda,mu]\n";
	gi.stopReadingCommandFile();
      }
      else
      {
	add( userDefined,"material interface" );
	printF("Material interface was chosen.\n");
      }
    }
    else if( answer=="stressNorm" )
    {
      add( userDefined,"stressNorm" );
      printF("Norm of the stress tensor was added.\n");
    }
    else if( answer=="shearStress" )
    {
      add( userDefined,"shearStress" );
      printF("The shear stress was added.\n");
    }
    else
    {
      printF("ERROR: unknown answer=%s\n",(const char*)answer);
      gi.stopReadingCommandFile();
    }
    
  }
  
  delete [] cNames;
  delete [] speciesNames;
  return 0;
}


//\begin{>>DerivedFunctionsInclude.tex}{\subsection{getUserDefinedDerivedFunction}} 
int DerivedFunctions:: 
getUserDefinedDerivedFunction( int index, 
                               int indexOut, 
                               const aString & name, 
                               const int numberOfComponents,
                               realCompositeGridFunction & uIn,
                               realCompositeGridFunction & uOut,
                               bool & interpolationRequired )
//==============================================================================================
// /Description:
//    Assign a user defined derived function. 
//
// /index (input): the index of this derived function (derived functions have index=0,1,2,...)
// /indexOut (input): fill in this component of uOut (indexOut=index+numberOfComponents)
// /name (input): the name associated with the derived function
// /uIn : the solution read from the show file
// /uOut : fill the derived function into this composite grid function
// /Return values: 0=success, non-zero=failure.
//\end{DerivedFunctionsInclude.tex}
//==============================================================================================
{
  assert( showFileReader!=NULL );
  ShowFileReader & reader = *showFileReader;


  CompositeGrid & cg = *uIn.getCompositeGrid();
  const int numberOfDimensions = cg.numberOfDimensions();
  
  const int nc = numberOfComponents;

  Index I1,I2,I3;

  interpolationRequired=false;  // set to true if the function should be interpolated before plotting

  if( name=="reaction rate" ) // this should match the "name" given in the "add" statement in the setup routine
  {
    // PDE parameters are saved in the OverBlown file OB_Parameters.C -- function saveParametersToShowFile()

    // get component numbers, e.g. density is uIn(I1,I2,I3,rc)
    int rc=-1, tc=-1, pc=-1, uc=-1, vc=-1, wc=-1, sc=-1;
    reader.getGeneralParameter("densityComponent",rc);   
    reader.getGeneralParameter("speciesComponent",sc);
    reader.getGeneralParameter("temperatureComponent",tc);
    reader.getGeneralParameter("pressureComponent",pc);
    reader.getGeneralParameter("uComponent",uc);
    reader.getGeneralParameter("vComponent",vc);
    reader.getGeneralParameter("wComponent",wc);

    bool ok;

    // reactionName: 
    //    "noReactions" 
    //    "onestep"
    //    "branching"
    //    "ignitionAndGrowth" 
    //    "oneEquationMixtureFraction"
    //    "twoEquationMixtureFractionAndExtentOfReaction"
    //    "chemkinReaction" 
    //    "unknown reactionType"

    aString reactionName;
    ok=reader.getGeneralParameter("reactionType",reactionName);  // ok==false if not found

    // For oneStep reaction:
    real heatRelease=0.,reciprocalActivationEnergy=0.,rateConstant=0.;
    ok=reader.getGeneralParameter("heatRelease",heatRelease);  // ok==false if not found
    ok=reader.getGeneralParameter("reciprocalActivationEnergy",reciprocalActivationEnergy);
    ok=reader.getGeneralParameter("rateConstant",rateConstant);

    // for chain branching
    real reciprocalActivationEnergyI, reciprocalActivationEnergyB, crossOverTemperatureI, 
      crossOverTemperatureB, absorbedEnergy;
    
    reader.getGeneralParameter("reciprocalActivationEnergyI",reciprocalActivationEnergyI);
    reader.getGeneralParameter("reciprocalActivationEnergyB",reciprocalActivationEnergyB);
    reader.getGeneralParameter("crossOverTemperatureI",crossOverTemperatureI);
    reader.getGeneralParameter("crossOverTemperatureB",crossOverTemperatureB);
    reader.getGeneralParameter("absorbedEnergy",absorbedEnergy);

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);
        
      int species=derived(index,1);  // here is the integer parameter we saved with the derived function
      assert( species>=0 && species<numberOfComponents );
      // density:  vIn(I1,I2,I3,rc); 
      // pressure: vIn(I1,I2,I3,pc);

      // ak1=reaction rate for chain-initiation
      // ak2=reaction rate for chain-branching
      //     ak1=pr(1)*dexp(-at(1)/temp)
      //     ak2=pr(2)*dexp(-at(2)/temp)

      vOut(I1,I2,I3,indexOut) = vIn(I1,I2,I3,species);   // do this for now

    }
  }
  else if( name=="gas mass fraction" )
  {
    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);

      vOut(I1,I2,I3,indexOut) = (1.-vIn(I1,I2,I3,8))*vIn(I1,I2,I3,4)/((1.-vIn(I1,I2,I3,8))*vIn(I1,I2,I3,4)+vIn(I1,I2,I3,8)*vIn(I1,I2,I3,0));
    }
  }
  else if( name=="mixture density" )
  {
    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);

      vOut(I1,I2,I3,indexOut) = (1.-vIn(I1,I2,I3,8))*vIn(I1,I2,I3,4)+vIn(I1,I2,I3,8)*vIn(I1,I2,I3,0);
    }
  }
  else if( name=="mixture schlieren" )
  {
    // we should interpolate this quantity since we can't compute it directly at interpolation points:
    interpolationRequired=false;
    real sMin=REAL_MAX;
    real sMax=0.;

    CompositeGrid & gc = *uIn.getCompositeGrid();
    CompositeGridOperators cgop(gc);

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      MappedGridOperators & op = cgop[grid];
    
      #ifdef USE_PPP
        realSerialArray vIn;  getLocalArrayWithGhostBoundaries(uIn[grid],vIn);
        realSerialArray vOut; getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
        intSerialArray mask;  getLocalArrayWithGhostBoundaries(mg.mask(),mask);
      #else
        const realSerialArray & vIn = uIn[grid];
        const realSerialArray & vOut = uOut[grid];
        const intSerialArray & mask = mg.mask();
      #endif

      getIndex(mg.dimension(),I1,I2,I3);
      bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
      if( !ok ) continue; // no points on this processor

// assuming 2d flow
      realSerialArray rsx(I1,I2,I3), rsy(I1,I2,I3), asx(I1,I2,I3), asy(I1,I2,I3), rgx(I1,I2,I3), rgy(I1,I2,I3);

      rsx=0.; rsy=0.; asx=0.; asy=0.; rgx=0.; rgy=0.;

      op.derivative(MappedGridOperators::xDerivative,vIn,rsx,I1,I2,I3,0);
      op.derivative(MappedGridOperators::yDerivative,vIn,rsy,I1,I2,I3,0);
      op.derivative(MappedGridOperators::xDerivative,vIn,asx,I1,I2,I3,8);
      op.derivative(MappedGridOperators::yDerivative,vIn,asy,I1,I2,I3,8);
      op.derivative(MappedGridOperators::xDerivative,vIn,rgx,I1,I2,I3,4);
      op.derivative(MappedGridOperators::yDerivative,vIn,rgy,I1,I2,I3,4);

      vOut(I1,I2,I3,indexOut)=sqrt( SQR((1-vIn(I1,I2,I3,8))*rgx(I1,I2,I3)+vIn(I1,I2,I3,8)*rsx(I1,I2,I3)+(vIn(I1,I2,I3,0)-vIn(I1,I2,I3,4))*asx(I1,I2,I3)) + SQR((1-vIn(I1,I2,I3,8))*rgy(I1,I2,I3)+vIn(I1,I2,I3,8)*rsy(I1,I2,I3)+(vIn(I1,I2,I3,0)-vIn(I1,I2,I3,4))*asy(I1,I2,I3)) );

//      where( mask(I1,I2,I3)<=0 )
//        vOut(I1,I2,I3,indexOut)=0.;

//      sMin=min(sMin,min(vOut(I1,I2,I3,indexOut)));
//      sMax=max(sMax,max(vOut(I1,I2,I3,indexOut)));

      Index J1,J2,J3;
      getIndex(mg.indexRange(),J1,J2,J3);
      ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,J1,J2,J3);
      if( ok )
      {
        where( mask(J1,J2,J3)>0 )
        {
          sMin=min(sMin,min(vOut(J1,J2,J3,indexOut)));
          sMax=max(sMax,max(vOut(J1,J2,J3,indexOut)));
        }
      }

    }
      printF("sMin sMax = %e %e\n",sMin,sMax);

//    #ifdef USE_PPP
//     sMin=ParallelUtility::getMinValue(sMin);
//     sMax=ParallelUtility::getMaxValue(sMax);
//    #endif


//    sMax=300.;
//    sMin=0.;

    if( fabs(sMax-sMin)>0. )
    {
      real exposure=1.;
      real amplification=15.;
      for( int grid=0; grid<cg.numberOfGrids(); grid++ )
      {
        #ifdef USE_PPP
          realSerialArray vOut;  getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
        #else
          realSerialArray & vOut = uOut[grid];
        #endif

        getIndex(cg[grid].dimension(),I1,I2,I3);
        bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
        if( ok )
        {
          vOut(I1,I2,I3,indexOut)=(vOut(I1,I2,I3,indexOut)-sMin)*(1./(sMax-sMin));
          vOut(I1,I2,I3,indexOut)=exposure*exp(-amplification*vOut(I1,I2,I3,indexOut));
        }
      }
    }

  }
  else if( name=="scaled schlieren" )
  {
    // we should interpolate this quantity since we can't compute it directly at interpolation points:
    interpolationRequired=false;
    real sMin=REAL_MAX;
    real sMax=0.;

    real rhoSolid=2297.;
    real rhoGas=1.4;
    real gamSolid=5.;
    real gamGas=1.4;

    real mu1Solid=1./(gamSolid-1.);
    real mu1Gas=1./(gamGas-1.);

    CompositeGrid & gc = *uIn.getCompositeGrid();
    CompositeGridOperators cgop(gc);

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      MappedGridOperators & op = cgop[grid];
    
      #ifdef USE_PPP
        realSerialArray vIn;  getLocalArrayWithGhostBoundaries(uIn[grid],vIn);
        realSerialArray vOut; getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
        intSerialArray mask;  getLocalArrayWithGhostBoundaries(mg.mask(),mask);
      #else
        const realSerialArray & vIn = uIn[grid];
        const realSerialArray & vOut = uOut[grid];
        const intSerialArray & mask = mg.mask();
      #endif

      getIndex(mg.dimension(),I1,I2,I3);
      bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
      if( !ok ) continue; // no points on this processor

// assuming 2d flow
      realSerialArray rx(I1,I2,I3), ry(I1,I2,I3), lambda(I1,I2,I3), scale(I1,I2,I3);

      rx=0.; ry=0.; lambda=0.; scale=1.;

      op.derivative(MappedGridOperators::xDerivative,vIn,rx,I1,I2,I3,0);
      op.derivative(MappedGridOperators::yDerivative,vIn,ry,I1,I2,I3,0);

      lambda(I1,I2,I3)=(vIn(I1,I2,I3,4)-mu1Gas)*(1./(mu1Solid-mu1Gas));
      scale(I1,I2,I3)=rhoSolid*lambda(I1,I2,I3)+rhoGas*(1-lambda(I1,I2,I3));

      vOut(I1,I2,I3,indexOut)=sqrt(SQR(rx(I1,I2,I3))+SQR(ry(I1,I2,I3)))/scale(I1,I2,I3);

//      where( mask(I1,I2,I3)<=0 )
//        vOut(I1,I2,I3,indexOut)=0.;

//      sMin=min(sMin,min(vOut(I1,I2,I3,indexOut)));
//      sMax=max(sMax,max(vOut(I1,I2,I3,indexOut)));

      Index J1,J2,J3;
      getIndex(mg.indexRange(),J1,J2,J3);
      ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,J1,J2,J3);
      if( ok )
      {
        where( mask(J1,J2,J3)>0 )
        {
          sMin=min(sMin,min(vOut(J1,J2,J3,indexOut)));
          sMax=max(sMax,max(vOut(J1,J2,J3,indexOut)));
        }
      }

    }
      printF("sMin sMax = %e %e\n",sMin,sMax);

//    #ifdef USE_PPP
//     sMin=ParallelUtility::getMinValue(sMin);
//     sMax=ParallelUtility::getMaxValue(sMax);
//    #endif


//    sMax=300.;
//    sMin=0.;

    if( fabs(sMax-sMin)>0. )
    {
      real exposure=1.;
      real amplification=45.;
      for( int grid=0; grid<cg.numberOfGrids(); grid++ )
      {
        #ifdef USE_PPP
          realSerialArray vOut;  getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
        #else
          realSerialArray & vOut = uOut[grid];
        #endif

        getIndex(cg[grid].dimension(),I1,I2,I3);
        bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
        if( ok )
        {
          vOut(I1,I2,I3,indexOut)=(vOut(I1,I2,I3,indexOut)-sMin)*(1./(sMax-sMin));
          vOut(I1,I2,I3,indexOut)=exposure*exp(-amplification*vOut(I1,I2,I3,indexOut));
        }
      }
    }

  }
  else if( name=="mixture temperature" )
  {
    real p0=1.;
    real rhoGas=1.4;
    real cvSolid=1.5;
    real cvGas=0.717625;
    real gamSolid=5.;
    real gamGas=1.4;

    real mu1Solid=1./(gamSolid-1.);
    real mu1Gas=1./(gamGas-1.);
    real T0=p0*mu1Gas/(rhoGas*cvGas);

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);

      vOut(I1,I2,I3,indexOut) = (vIn(I1,I2,I3,4)-mu1Gas)*(1./(mu1Solid-mu1Gas));  // lambda
      vOut(I1,I2,I3,indexOut) = cvSolid*vOut(I1,I2,I3,indexOut)+cvGas*(1.-vOut(I1,I2,I3,indexOut));  // mixture Cv
      vOut(I1,I2,I3,indexOut) = (vIn(I1,I2,I3,3)*vIn(I1,I2,I3,4)+vIn(I1,I2,I3,5)/vIn(I1,I2,I3,0))/(T0*vOut(I1,I2,I3,indexOut));  // mixture temperature
    }
  }
  else if( name=="masked progress" )
  {
    int sc;
    real cutOff = 0.5;
    double notANumber=10.0;

    reader.getGeneralParameter("speciesComponent",sc);
    if( sc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: speciesComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }
    
    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);

      where( vIn(I1,I2,I3,sc) >= cutOff )
      {
	vOut(I1,I2,I3,indexOut) = vIn(I1,I2,I3,sc+1);
      }
      where( vIn(I1,I2,I3,sc) < cutOff )
      {
	vOut(I1,I2,I3,indexOut) = notANumber;
      }
    }
  }
  else if( name=="material interface" )
  {
    CompositeGrid & gc = *uIn.getCompositeGrid();
    CompositeGridOperators op(gc);
    uIn.setOperators(op);
    int sc;

    reader.getGeneralParameter("speciesComponent",sc);
    if( sc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: speciesComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      realMappedGridFunction & vIn = uIn[grid];
      realMappedGridFunction & vOut = uOut[grid];

      getIndex(mg.dimension(),I1,I2,I3);
      vOut(I1,I2,I3,indexOut) = SQRT(SQR(vIn.x()(I1,I2,I3,sc))+SQR(vIn.y()(I1,I2,I3,sc)));
      vOut(I1,I2,I3,indexOut) = 1.0/(1.0+vOut(I1,I2,I3,indexOut));
    }
  }
  else if( name=="stressNorm" )
  {
    // we should interpolate this quantity since we can't compute it directly at interpolation points:
    interpolationRequired=true;  

    CompositeGrid & gc = *uIn.getCompositeGrid();
    CompositeGridOperators cgop(gc);
    int sc;

    int uc=-1, vc=-1, wc=-1;
    reader.getGeneralParameter("uComponent",uc);
    reader.getGeneralParameter("vComponent",vc);
    reader.getGeneralParameter("wComponent",wc);

    if( uc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: uComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }
    if( vc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: vComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }
    if( cg.numberOfDimensions()==3 && wc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: wComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }

    real nu=0.;
    reader.getGeneralParameter("nu",nu);
    

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      MappedGridOperators & op = cgop[grid];
    
      #ifdef USE_PPP
        realSerialArray vIn;  getLocalArrayWithGhostBoundaries(uIn[grid],vIn);
        realSerialArray vOut; getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
      #else
        const realSerialArray & vIn = uIn[grid];
        const realSerialArray & vOut = uOut[grid];
      #endif

      getIndex(mg.dimension(),I1,I2,I3);
      bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
      if( !ok ) continue; // no points on this processor

      if( mg.numberOfDimensions()==2 )
      {
        realSerialArray ux(I1,I2,I3), uy(I1,I2,I3), vx(I1,I2,I3), vy(I1,I2,I3);

        op.derivative(MappedGridOperators::xDerivative,vIn,ux,I1,I2,I3,uc);	
        op.derivative(MappedGridOperators::yDerivative,vIn,uy,I1,I2,I3,uc);	
        op.derivative(MappedGridOperators::xDerivative,vIn,vx,I1,I2,I3,vc);	
        op.derivative(MappedGridOperators::yDerivative,vIn,vy,I1,I2,I3,vc);	

        // check this: norm of the INS stress tensor: 
        vOut(I1,I2,I3,indexOut)= nu*sqrt( SQR(2.*ux) + 2.*SQR(uy+vx) + SQR(2.*vy) );

      }
      else if( mg.numberOfDimensions()==3 )
      {
        realSerialArray ux(I1,I2,I3), uy(I1,I2,I3), uz(I1,I2,I3);
        realSerialArray vx(I1,I2,I3), vy(I1,I2,I3), vz(I1,I2,I3);
        realSerialArray wx(I1,I2,I3), wy(I1,I2,I3), wz(I1,I2,I3);

        op.derivative(MappedGridOperators::xDerivative,vIn,ux,I1,I2,I3,uc);	
        op.derivative(MappedGridOperators::yDerivative,vIn,uy,I1,I2,I3,uc);	
        op.derivative(MappedGridOperators::zDerivative,vIn,uz,I1,I2,I3,uc);	

        op.derivative(MappedGridOperators::xDerivative,vIn,vx,I1,I2,I3,vc);	
        op.derivative(MappedGridOperators::yDerivative,vIn,vy,I1,I2,I3,vc);	
        op.derivative(MappedGridOperators::zDerivative,vIn,vz,I1,I2,I3,vc);	

        op.derivative(MappedGridOperators::xDerivative,vIn,wx,I1,I2,I3,wc);	
        op.derivative(MappedGridOperators::yDerivative,vIn,wy,I1,I2,I3,wc);	
        op.derivative(MappedGridOperators::zDerivative,vIn,wz,I1,I2,I3,wc);	


        // check this: norm of the INS stress tensor: 
        vOut(I1,I2,I3,indexOut)= nu*sqrt( SQR(2.*ux) + 2.*SQR(uy+vx) + 2.*SQR(uz+wx) + 
                                          SQR(2.*vy) + 2.*SQR(vz+wy) + 
                                          SQR(wz) );
      }
      else
      {
        realSerialArray ux(I1,I2,I3);
        op.derivative(MappedGridOperators::xDerivative,vIn,ux,I1,I2,I3,uc);	

        // check this: norm of the INS stress tensor: 
        vOut(I1,I2,I3,indexOut)= nu*sqrt( SQR(2.*ux) );
      }
    }
  }
  else if( name=="shearStress" )
  {
    // --- compute the shear stress on boundaries ---

    // we should interpolate this quantity since we can't compute it directly at interpolation points:
    interpolationRequired=true;  

    CompositeGrid & gc = *uIn.getCompositeGrid();
    CompositeGridOperators cgop(gc);
    int sc;

    int uc=-1, vc=-1, wc=-1;
    reader.getGeneralParameter("uComponent",uc);
    reader.getGeneralParameter("vComponent",vc);
    reader.getGeneralParameter("wComponent",wc);
    if( uc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: uComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }
    if( vc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: vComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }
    if( cg.numberOfDimensions()==3 && wc<0 )
    {
      printF("getUserDefinedDerivedFunction:%s:ERROR: wComponent is not in the show file!\n",
               (const char*)name);
      return 1;
    }

    real nu=0.;
    reader.getGeneralParameter("nu",nu);
    

    for( int grid=0; grid<cg.numberOfGrids(); grid++ )
    {
      MappedGrid & mg = cg[grid];
      MappedGridOperators & op = cgop[grid];
      mg.update(MappedGrid::THEvertexBoundaryNormal); // make sure the normals are there
    
      #ifdef USE_PPP
        realSerialArray vIn;  getLocalArrayWithGhostBoundaries(uIn[grid],vIn);
        realSerialArray vOut; getLocalArrayWithGhostBoundaries(uOut[grid],vOut);
      #else
        const realSerialArray & vIn = uIn[grid];
        const realSerialArray & vOut = uOut[grid];
      #endif

      Range all;
      vOut(all,all,all,indexOut)=0.;
	

      for( int side=0; side<=1; side++ )for( int axis=0; axis<mg.numberOfDimensions(); axis++ )
      {
	if( mg.boundaryCondition(side,axis)>0 )
	{
          // --- this is a physical boundary ---

#ifdef USE_PPP
	  const realSerialArray & normal  = mg.vertexBoundaryNormalArray(side,axis);
#else
	  const realSerialArray & normal  = mg.vertexBoundaryNormal(side,axis);
#endif

	  getBoundaryIndex(mg.gridIndexRange(),side,axis,I1,I2,I3);

	  bool ok = ParallelUtility::getLocalArrayBounds(uOut[grid],vOut,I1,I2,I3);
	  if( !ok ) continue; // no points on this processor

	  if( mg.numberOfDimensions()==2 )
	  {
	    realSerialArray ux(I1,I2,I3), uy(I1,I2,I3), vx(I1,I2,I3), vy(I1,I2,I3);

	    op.derivative(MappedGridOperators::xDerivative,vIn,ux,I1,I2,I3,uc);	
	    op.derivative(MappedGridOperators::yDerivative,vIn,uy,I1,I2,I3,uc);	
	    op.derivative(MappedGridOperators::xDerivative,vIn,vx,I1,I2,I3,vc);	
	    op.derivative(MappedGridOperators::yDerivative,vIn,vy,I1,I2,I3,vc);	

            // The normal traction vector is n.sigma 
            //  T_n = ( n1*s_11 + n2*s_12 , n1*s_21 + n2*s_22 )
            //      = force on the boundary
            //  shear-stress = T_n - (n.T_n)n    (subtract off normal component of the traction)
            //              
            realSerialArray t1(I1,I2,I3), t2(I1,I2,I3), nDotT(I1,I2,I3);
            // no need to include the pressure :
            t1 = normal(I1,I2,I3,0)*( 2.*ux ) + normal(I1,I2,I3,1)*( uy+vx );
            t2 = normal(I1,I2,I3,0)*( uy+vx ) + normal(I1,I2,I3,1)*( 2.*vy );
	    
            nDotT= normal(I1,I2,I3,0)*t1 + normal(I1,I2,I3,1)*t2;

            t1 -= nDotT*normal(I1,I2,I3,0);
            t2 -= nDotT*normal(I1,I2,I3,1);

	    // norm of the shear stress vector
	    vOut(I1,I2,I3,indexOut)= nu*sqrt( SQR(t1) + SQR(t2) );

	  }
	  else if( mg.numberOfDimensions()==3 )
	  {
	    realSerialArray ux(I1,I2,I3), uy(I1,I2,I3), uz(I1,I2,I3);
	    realSerialArray vx(I1,I2,I3), vy(I1,I2,I3), vz(I1,I2,I3);
	    realSerialArray wx(I1,I2,I3), wy(I1,I2,I3), wz(I1,I2,I3);

	    op.derivative(MappedGridOperators::xDerivative,vIn,ux,I1,I2,I3,uc);	
	    op.derivative(MappedGridOperators::yDerivative,vIn,uy,I1,I2,I3,uc);	
	    op.derivative(MappedGridOperators::zDerivative,vIn,uz,I1,I2,I3,uc);	

	    op.derivative(MappedGridOperators::xDerivative,vIn,vx,I1,I2,I3,vc);	
	    op.derivative(MappedGridOperators::yDerivative,vIn,vy,I1,I2,I3,vc);	
	    op.derivative(MappedGridOperators::zDerivative,vIn,vz,I1,I2,I3,vc);	

	    op.derivative(MappedGridOperators::xDerivative,vIn,wx,I1,I2,I3,wc);	
	    op.derivative(MappedGridOperators::yDerivative,vIn,wy,I1,I2,I3,wc);	
	    op.derivative(MappedGridOperators::zDerivative,vIn,wz,I1,I2,I3,wc);	


            // The normal stress is n.sigma 
            //  normal-stress = ( n1*s_11 + n2*s_12 + n3*s_13, n1*s_21 + n2*s_22 + n3*s_23,n1*s_31 + n2*s_32 + n3*s_33)

            // The normal traction vector is n.sigma 
            //  T_n = ( n1*s_11 + n2*s_12 + n3*s_13, n1*s_21 + n2*s_22 + n3*s_23,n1*s_31 + n2*s_32 + n3*s_33)
            //      = force on the boundary
            //  shear-stress = T_n - (n.T_n)n    (subtract off normal component of the traction)
            //              
            realSerialArray t1(I1,I2,I3), t2(I1,I2,I3), t3(I1,I2,I3), nDotT(I1,I2,I3);
            // no need to include the pressure :
            t1 = normal(I1,I2,I3,0)*( 2.*ux ) + normal(I1,I2,I3,1)*( uy+vx ) + normal(I1,I2,I3,2)*( uz+wx );
            t2 = normal(I1,I2,I3,0)*( uy+vx ) + normal(I1,I2,I3,1)*( 2.*vy ) + normal(I1,I2,I3,2)*( vz+wy );
            t3 = normal(I1,I2,I3,0)*( uz+wx ) + normal(I1,I2,I3,1)*( vz+wy ) + normal(I1,I2,I3,2)*( 2.*wz );
	    
            nDotT= normal(I1,I2,I3,0)*t1 + normal(I1,I2,I3,1)*t2 + normal(I1,I2,I3,2)*t3;

            t1 -= nDotT*normal(I1,I2,I3,0);
            t2 -= nDotT*normal(I1,I2,I3,1);
            t3 -= nDotT*normal(I1,I2,I3,2);

	    // norm of the shear stress vector
	    vOut(I1,I2,I3,indexOut)= nu*sqrt( SQR(t1) + SQR(t2) + SQR(t3) );
	  }
	  else
	  {
	    vOut(I1,I2,I3,indexOut)= 0.;
	  }
	}
      } // end for side, for axis 
      
    }
  }
  else
  {
    printF("DerivedFunctions::getUserDefinedDerivedFunction: ERROR: unknown user defined function: %s\n",
	   (const char*)name);
    return 1;
  }
  

  return 0;
}

